import asyncio
import aiohttp
import aiofiles
import json
import codecs
from aiohttp import web
import aiohttp_cors
from traceback import print_exc
from appPublic.sshx import SSHNode
from .baseProcessor import BaseProcessor, PythonScriptProcessor

class XtermProcessor(BaseProcessor):
	@classmethod
	def isMe(self,name):
		return name=='xterm'

	async def ws_2_process(self, ws):
		async for msg in ws:
			if msg.type == aiohttp.WSMsgType.TEXT:
				self.p_obj.stdin.write(msg.data)
			elif msg.type == aiohttp.WSMsgType.ERROR:
				# print('ws connection closed with exception %s' % ws.exception())
				return
		
	async def process_2_ws(self, ws):
		while self.running:
			x = await self.p_obj.stdout.read(1024)
			await self.ws_sendstr(ws, x)

	async def datahandle(self,request):
		await self.path_call(request)
		
	async def path_call(self, request, params={}):
		await self.set_run_env(request)
		lenv = self.run_ns.copy()
		lenv.update(params)
		del lenv['request']
		ws = web.WebSocketResponse()
		await ws.prepare(request)
		await self.create_process()
		self.ws_sendstr(ws, 'Welcom to sshclient')
		r1 = self.ws_2_process(ws)
		r2 = self.process_2_ws(ws)
		await asyncio.gather(r1,r2)
		self.retResponse = ws
		return ws

	async def get_login_info(self):
		async with aiofiles.open(self.real_path, 'r', encoding='utf-8') as f:
			txt = await f.read()
			self.login_info = json.loads(txt)
		# print(f'{self.login_info=}')

	async def create_process(self):
		# id = lenv['params_kw'].get('termid')
		await self.get_login_info()
		host = self.login_info['host']
		port = self.login_info.get('port', 22)
		username = self.login_info.get('username', 'root')
		password = self.login_info.get('password',None)
		self.sshnode = SSHNode(host, username=username,
									password=password,
									port=port)
		await self.sshnode.connect()
		self.p_obj = await self.sshnode._process('bash',
											term_type='vt100',
											term_size=(80, 24),
											encoding='utf-8')
		self.running = True

	async def ws_sendstr(self, ws:web.WebSocketResponse, s:str):
		data = {
			"type":1,
			"data":s
		}
		await ws.send_str(json.dumps(data))

	def close_process(self):
		self.sshnode.close()
		self.p_obj.close()

async def ws_send(ws:web.WebSocketResponse, data):
	if not isinstance(data, str):
		data = json.dumps(data)
	d = {
		"type":1,
		"data":data
	}
	d = json.dumps(d)
	try:
		return await ws.send_str(d)
	except Exception as e:
		print('ws.send_str() error:', e)
		return False

class WsPool:
	def __init__(self, ws, ws_path, app):
		self.app = app
		self.id = None
		self.ws = ws
		print(f'-----------{self.id=}-----------')
		self.ws_path = ws_path

	def get_data(self):
		return self.app.get_data(self.ws_path)

	def set_data(self, data):
		self.app.set_data(self.ws_path, data)

	def is_online(self, userid):
		data = self.get_data()
		ws = data.get(userid)
		if ws is None:
			return False
		return True

	def register(self, id):
		self.id = id
		self.add_mine()

	def add_mine(self):
		data = self.get_data()
		if data is None:
			data = {}
		data.update({self.id:self.ws})
		self.set_data(data)

	def delete_mine(self):
		data = self.get_data()
		if not data.get(self.id):
			return
		data = {k:v for k,v in self.data.items() if k != self.id }
		self.set_data(data)
		
	async def sendto(self, data, id=None):
		if id is None:
			return await ws_send(self.ws, data)
		d = self.get_data()
		ws = d.get(id)
		return await ws_send(ws, data)

class WebsocketProcessor(PythonScriptProcessor):
	@classmethod
	def isMe(self,name):
		return name=='ws'

	async def path_call(self, request,params={}):
		await self.set_run_env(request)
		lenv = self.run_ns.copy()
		lenv.update(params)
		params_kw = lenv.params_kw
		userid = lenv.params_kw.userid or await lenv.get_user()
		del lenv['request']
		txt = await self.loadScript(self.real_path)
		ws = web.WebSocketResponse()
		try:
			await ws.prepare(request)
		except Exception as e:
			print('--------except:', e, type(e))
			print_exc()
			raise e
		ws_pool = WsPool(ws, request.path, request.app)
		async for msg in ws:
			if msg.type == aiohttp.WSMsgType.TEXT:
				if msg.data == 'exit':
					break
				lenv['ws_data'] = msg.data
				lenv['ws_pool'] = ws_pool
				exec(txt,lenv,lenv)
				func = lenv['myfunc']
				resp =  await func(request,**lenv)
				# await self.ws_sendstr(ws, resp)
			elif msg.type == aiohttp.WSMsgType.ERROR:
				# print('ws connection closed with exception %s' % ws.exception())
				break
			else:
				print('datatype error', msg.type)
		ws_pool.delete_mine()
		self.retResponse =  ws
		await ws.close()
		return ws

